
open Ast
open Ast_types
open Environment
open Equations_types

type envType = Environment_types.t
type typedefs = typedef list

(* global flag *)
let type_inference = ref true

(* Gets the return type of a functional type 
For example, getReturnType (a -> b -> c -> d) 1 returns b -> c -> d 
getReturnType (a -> b -> c -> d) 2 returns c -> d 
getReturnType (a -> b -> c -> d) 3 returns d *)
let rec getReturnType (t : typeNC) (n : int) =
  if n <= 0 then t else
  match t with
    | Arrow (_, b) -> getReturnType b (n-1)
    | _ -> t
  
(* alpha-convert a schema *)
(* This function returns the type part of s in which every quantified 
 variable is replaced by a fresh variable. For example, if s is 
 "for all a,b , a -> b -> Integer", this function returns 
 "[c;d], c -> d -> Integer", where c and d are fresh variables. *)
let alpha (Schema (vars, typ) : schema) : id list * typeNC =
  List.fold_right (fun x (freshvars, t) ->
    let y = newId() in
    y :: freshvars, substituteExp (VarType x) (VarType y) t
  ) vars ([], typ)
                                                                      
let rec type_constructor (c : id) (td : typedef list) 
  : id * id list * typeNC option = match td with
(* Given a type constructor c, finds out the corresponding type id,
   with its alpha-converted type variables, and the argument to 
   constructor c containing those type variables *)
  | [ ] -> raise (Type_error ("Constructor " ^ c ^ " undefined"))
  | (id, (typl, l)) :: tl ->
    try match List.assoc c l with
      None -> (id, List.map (fun x -> newId()) typl, None)
    | Some(argtyps) -> let typl1, typ1 = alpha(Schema(typl, argtyps)) in
                       (id, typl1, Some(typ1))
    with Not_found -> type_constructor c tl

let rec subst_type (t : typeNC) (a : (idtype * typeNC) list) = match t with
| VarType i -> (try List.assoc i a with Not_found -> VarType i)
| Integer | FloatType | String | Boolean | UnitType | DummyType | Null -> t
| Arrow (t1, t2) -> Arrow (subst_type t1 a, subst_type t2 a)
| TupleType (tl) -> TupleType (List.map (fun t1 -> subst_type t1 a) tl)
| UserType (s, tl) ->
    UserType (s, List.map (fun t1 -> subst_type t1 a) tl)

let rec gts_pattern (p : pattern) (o : envType * typedefs)
                   : typeNC * (envType * typedefs) * substitution =
    match p with
    | PVar x -> (* it is binding a new variable x! *)
        let freshVar = newVar() in
        let typ = Schema ([], freshVar) in
        freshVar, ((Environment_types.bind x typ (fst o)), snd o), []
    (* just slightly adapted for the gts calls *)
    | PInt i -> (Integer, o, [])
    | PFloat f -> (FloatType, o, [])
    | PStrg s -> (String, o, [])
    | PBool b -> (Boolean, o, [])
    | PUnit -> (UnitType, o, [])
    | PTuple l ->
      let rec iterate oi = function
        | [ ] -> [ ], oi, [ ]
        | ei :: tl ->
          let (ti, oi1, substi) = gts_pattern ei oi in 
          let ri = solve substi in
          let oj = applySubstEnv ri (fst oi1) in
          let (lj, oi2, substj) = iterate (oj, snd oi1) tl in
          let rj = solve substj in
          (applySubst rj ti) :: lj, oi2, ri @ rj in
      let l2, o2, subst2 = iterate o l in
      TupleType l2, o2, subst2
    | PInj(c, eop) ->
      let (idtyp, typvars, typop) = type_constructor c (snd o) in
        (match eop, typop with
         | None, None -> 
           UserType(idtyp, List.map (fun x -> VarType x) typvars), o, [ ]
         | _, None -> 
           raise (Type_error (
	     "Constructor " ^ c ^ " does not expect an argument"))
         | None, _ -> 
           raise (Type_error (
	     "Constructor " ^ c ^ " expects an argument"))
         | Some(e1), Some(typop1) -> (* just check that e1 has type t *)
           let (typ1, o1, subst1) = gts_pattern e1 o in
           let ro = solve subst1 in
           let sigma = solve [(applySubst ro typ1, typop1)] in
           let typ0 = UserType(idtyp, List.map (fun x -> VarType x) typvars) in
           (applySubst sigma typ0, o1, ro @ sigma)
        )  

and gts_patterns (pl : (pattern * expr) list) (o : envType * typedefs) :
    typeNC * typeNC * (envType * typedefs) * substitution =
  match pl with
  | [ ] -> raise (Type_error "Unexpected empty list of patterns")
  | (pi, ei) :: tl ->
      let (typ1, o1, subst1) = gts_pattern pi o in
      let r1 = solve subst1 in
      let (typ2, o2, subst2) = gts ei (applySubstEnv r1 (fst o1), snd o1) in
      let r2 = solve subst2 in
      let typ1' = applySubst r2 typ1 in
      let sigma = (
        if tl = [ ] then [ ] else
          let (typ3, typ4, o3, subst3) = gts_patterns tl o in
          solve [ (typ1', typ3); (typ2, typ4) ]) in
      (applySubst sigma typ1', applySubst sigma typ2, o, r1 @ r2 @ sigma)

(* gts = Get Type and Substitution 
This function takes an expression t and a type environment o and 
returns a triple: the expression's type if it has one, 
the new type environment, and the substitution solution of the type 
equations generated by type inference.
Raises Type_error if expression t is not well-typed. *)
and gts (t : expr) (o : envType * typedefs)
            : typeNC * (envType * typedefs) * substitution =
    match t with
      | Var x -> snd (alpha (Environment_types.lookup x (fst o))), o, []
      | Int i -> (Integer, o, [])
      | Float f -> (FloatType, o, [])
      | Strg s -> (String, o, [])
      | Bool b -> (Boolean, o, [])

      (* functions *)
      | Fun ([Var x], t) -> 
          let freshVar = newVar() in
          let (typ, _, subst) = 
            gts t ((Environment_types.bind x
		      (Schema ([], freshVar)) (fst o)), snd o) in
          let ro = solve subst in
          (applySubst ro (Arrow (freshVar, typ)), o, ro)
      | Fun (Var x :: tl, t) -> 
          let freshVar = newVar() in
          let (typ, _, subst) = 
            gts (Fun (tl, t)) 
                ((Environment_types.bind x 
		    (Schema ([], freshVar)) (fst o)), snd o) in
          let ro = solve subst in
          (applySubst ro (Arrow (freshVar, typ)), o, ro)
      | Fun ([Unit], t) -> 
          let (typ, _, subst) = gts t o in
          let ro = solve subst in
          (applySubst ro (Arrow (UnitType, typ)), o, ro)
      | Fun (Unit :: tl, t) -> 
          let (typ, _, subst) = gts (Fun (tl, t)) o in
          let ro = solve subst in
          (applySubst ro (Arrow (UnitType, typ)), o, ro)
      | Fun (_, t) -> type_error "Unknown function arguments"

      (* typed functions *)
      | FunType ([Var x], [Null], t) -> gts (Fun ([Var x], t)) o 
      | FunType ([Var x], [typex], t) -> 
          let (typ, _, subst) = 
            gts t ((Environment_types.bind x 
		      (Schema ([], typex)) (fst o)), snd o) in
          let ro = solve subst in
          (applySubst ro (Arrow (typex, typ)), o, ro)
      | FunType (Var x :: tl, Null :: ttl, t) -> 
          let freshVar = newVar()  in
          let (typ, _, subst) = 
            gts (FunType(tl, ttl,t))
                ((Environment_types.bind x
		    (Schema ([], freshVar)) (fst o)), snd o) in
          let ro = solve subst in
          (applySubst ro (Arrow (freshVar, typ)), o, ro)
      | FunType (Var x :: tl, typex :: ttl, t) -> 
          let (typ, _, subst) = 
            gts (FunType(tl, ttl, t))
                ((Environment_types.bind x
		    (Schema ([], typex)) (fst o)), snd o) in
          let ro = solve subst in
          (applySubst ro (Arrow (typex, typ)), o, ro)
      | FunType ([Unit], [(Null | UnitType)], t) -> gts (Fun ([Unit], t)) o
      | FunType (Unit :: tl, (Null | UnitType) :: ttl, t) ->
          let (typ, _, subst) = gts (FunType (tl, ttl, t)) o in
          let ro = solve subst in 
          (applySubst ro (Arrow (UnitType, typ)), o, ro)
      | FunType (Unit :: _, _, _) -> type_error "type of () must be unit"
      | FunType _ -> type_error "Unknown function arguments"
      | FunCorec _ -> 
	type_error "Direct declaration of corecursive function not allowed"

      | (Let (id, args, e1, e2) | LetType (id, _, args, _, e1, e2) |
        Letrec (id, args, e1, e2) | LetrecType (id, _, args, _, e1, e2) |
        Letcorec (_, id, args, e1, e2) | LetcorecType (_, id, _, args, _, e1, e2)) -> 
          let d =
            (match t with 
              | Let _ -> Def (id, args, e1)
              | LetType (id, typeid, args, typeargs, e1, e2) -> 
                  DefType (id, typeid, args, typeargs, e1)
              | Letrec _ -> Defrec (id, args, e1)
              | LetrecType (id, typeid, args, typeargs, e1, e2) -> 
                  DefrecType (id, typeid, args, typeargs, e1)
              | Letcorec(s, id, args, e1, e2) -> Defcorec (s, id, args, e1)
              | LetcorecType (s, id, typeid, args, typeargs, e1, e2) -> 
                  DefcorecType (s, id, typeid, args, typeargs, e1)
              | _ -> type_error "impossible case") in
          let (typeid, (oP, td), s) = gtsd d o in
          let ro = solve s in
          let oS = applySubstEnv ro oP in
          let gen = generalize typeid oS in
          let oT = Environment_types.bind id gen oS in
          let (typ, _, subst) = gts e2 (oT, td) in
          let roP = solve subst in
          (applySubst roP typ, o, ro @ roP)

      | App (t, u) -> 
          let (typ1, _, subst1) = gts u o in
          let ro = solve subst1 in 
          let (typ2, _, subst2) = gts t ((applySubstEnv ro (fst o)), snd o) in
          let roP = solve subst2 in
          let freshVar = newVar() in
          let sigma = solve [(typ2, Arrow (applySubst roP typ1, freshVar))] in 
          (applySubst sigma freshVar, o, ro @ roP @ sigma)     

      | If (t, u, v) ->
          let (typ1, _, subst1) = gts t o in
          let ro = solve subst1 in
          let sigma = solve [(typ1, Boolean)] in
          let oP = applySubstEnv (ro @ sigma) (fst o) in
          let (typ2, _, subst2) = gts u (oP, snd o) in
          let roP = solve subst2 in
          let (typ3, _, subst3) = gts v ((applySubstEnv roP oP), snd o) in
          let roS = solve subst3 in
          let sigmaP = solve [(applySubst roS typ2, typ3)] in
          (applySubst sigmaP typ3, o, ro @ sigma @ roP @ roS @ sigmaP)
   
      | While (t, u) ->
          let (typ1, _, subst1) = gts t o in  
          let ro = solve subst1 in
          let sigma = solve [(typ1, Boolean)] in 
          let oP = applySubstEnv (ro @ sigma) (fst o) in    
          let (typ2, _, subst2) = gts u (oP, snd o) in
          let roP = solve subst2 in
          let sigmaP = solve [(typ2, UnitType)] in
          (UnitType, o, ro @ sigma @ roP @ sigmaP)

      | For (x, e1, e2, e3) ->
          let (typ1, _, subst1) = gts e1 o in
          let ro = solve subst1 in
          let sigma = solve [(typ1, Integer)] in
          let oP = applySubstEnv (ro @ sigma) (fst o) in  
          let (typ2, _, subst2) = gts e2 (oP, snd o) in
          let roP = solve subst2 in
          let sigmaP = solve [(typ2, Integer)] in
          let oS = applySubstEnv (roP @ sigmaP) oP in
          let (typ3, _, subst3) = 
            gts e3 ((Environment_types.bind x
		       (Schema ([], Integer)) oS), snd o) in
          let roS = solve subst3 in
          let sigmaS = solve [(typ3, UnitType)] in
          (UnitType, o, ro @ sigma @ roP @ sigmaP @ roS @ sigmaS)
            
      | Assg (x, e) ->
          let (typ1, _, subst1) = gts (Var x) o in
          let ro = solve subst1 in
          let oP = applySubstEnv ro (fst o) in
          let (typ2, _, subst2) = gts e (oP, snd o) in
          let roP = solve subst2 in    
          let sigma = solve [(applySubst roP typ1, typ2)] in
          (UnitType, o, ro @ roP @ sigma)
             
      | Seq (e1, e2) ->
          let (typ1, _, subst1) = gts e1 o in
          let ro = solve subst1 in
          let sigma = solve [(typ1, UnitType)] in
          let oP = applySubstEnv (ro @ sigma) (fst o) in  
          let (typ2, _, subst2) = gts e2 (oP, snd o) in
          let roP = solve subst2 in
          (applySubst roP typ2, o, ro @ sigma @ roP)

      | Match (e1, pl) -> (* DummyType, o, [ ] *)
          let (typ1, _, subst1) = gts e1 o in
          let r1 = solve subst1 in
          let (typ2, typ3, _, subst2) = gts_patterns pl o in
          let r2 = solve subst2 in
          let sigma = solve [ (applySubst r1 typ1, applySubst r2 typ2) ] in
          (applySubst sigma typ3, o, r1 @ r2 @ sigma)

      | Tuple l ->
          let rec iterate oi = function
            | [ ] -> [ ], oi, [ ]
            | ei :: tl ->
              let (ti, _, substi) = gts ei oi in 
              let ri = solve substi in
              let oj = applySubstEnv ri (fst oi) in
              let (lj, _, substj) = iterate (oj, snd oi) tl in
              let rj = solve substj in
              ti :: lj, oi, ri @ rj in
          let l2, o2, subst2 = iterate o l in
          TupleType l2, o2, subst2

      | Inj(c, eop) ->
          let (idtyp, typvars, typop) = type_constructor c (snd o) in
          (match eop, typop with
           | None, None -> 
             UserType(idtyp, List.map (fun x -> VarType x) typvars), o, [ ]
           | _, None -> 
             raise (Type_error (
	       "Constructor " ^ c ^ " does not expect an argument"))
           | None, _ -> 
             raise (Type_error (
	       "Constructor " ^ c ^ " expects an argument"))
           | Some(e1), Some(typop1) ->
             let (typ1, _, subst1) = gts e1 o in
             let ro = solve subst1 in
             let sigma = solve [(typ1, typop1)] in
             let typ0 = UserType(idtyp, 
				 List.map (fun x -> VarType x) typvars) in
             (applySubst sigma typ0, o, ro @ sigma)
          )

      (* Comparison operators *)
      | (Eq (e1, e2) | Neq (e1, e2)) -> 
          let (typ1, _, subst1) = gts e1 o in
          let ro = solve subst1 in
          let oP = applySubstEnv ro (fst o) in
          let (typ2, _, subst2) = gts e2 (oP, snd o) in
          let roP = solve subst2 in
          let sigma = solve [(typ2, applySubst roP typ1)] in
          (match typ2 with
            | (Integer | Boolean | VarType _ | String | TupleType _ 
               | FloatType | UserType _) -> 
                (Boolean, o, ro @ roP @ sigma)
            | _ -> type_error "can only test booleans, lists, integers, or strings for equality")
      | (Lt (e1,e2) | Le (e1, e2) | Gt (e1, e2) |  Ge (e1, e2)) ->  
          let (typ1, _, subst1) = gts e1 o in
          let ro = solve subst1 in
          let oP = applySubstEnv ro (fst o) in
          let (typ2, _, subst2) = gts e2 (oP, snd o) in
          let roP = solve subst2 in
          let sigma = solve [(typ2, applySubst roP typ1)] in
          (match typ2 with
            | (Integer | Boolean | VarType _ | String) -> (Boolean, o, ro @ roP @ sigma)
            | _ -> type_error "can only compare ints to ints and bools to bools")
  
      (* Arithmetic operators on integers *)
      | (Plus (e1, e2) | Minus (e1, e2) | Mul (e1, e2) | Div (e1, e2) | Mod(e1, e2)) -> 
          let (typ1, _, subst1) = gts e2 o in
          let ro = solve subst1 in
          let sigma = solve [(typ1, Integer)] in
          let (typ2, _, subst2) = gts e1 ((applySubstEnv (ro @ sigma) (fst o)), snd o) in
          let roP = solve subst2 in
          let sigmaP = solve [(typ2, Integer)] in
          (Integer, o, ro @ sigma @ roP @ sigmaP) 

      (* Arithmetic operators on floats *)
      | (PlusF (e1, e2) | MinusF (e1, e2) | MulF (e1, e2) | DivF (e1, e2)) -> 
          let (typ1, _, subst1) = gts e2 o in
          let ro = solve subst1 in
          let sigma = solve [(typ1, FloatType)] in
          let (typ2, _, subst2) = gts e1 ((applySubstEnv (ro @ sigma) (fst o)), snd o) in
          let roP = solve subst2 in
          let sigmaP = solve [(typ2, FloatType)] in
          (FloatType, o, ro @ sigma @ roP @ sigmaP) 

      (* String operators *)
      | Concat (e1, e2) ->
          let (typ1, _, subst1) = gts e1 o in
          let ro = solve subst1 in
          let sigma = solve [(typ1, String)] in
          let oP = applySubstEnv (ro @ sigma) (fst o) in
          let (typ2, _, subst2) = gts e2 (oP, snd o) in
          let roP = solve subst2 in
          let sigmaP = solve [(typ2, String)] in 
          (String, o, ro @ sigma @ roP @ sigmaP)

      (* Boolean operators *)
      | Not e1 ->
          let (typ, _, subst) = gts e1 o in
          let ro = solve subst in
          let sigma = solve [(typ, Boolean)] in
          (Boolean, o, ro @ sigma)
      | (And (e1, e2) | Or (e1, e2)) ->
          let (typ1, _, subst1) = gts e2 o in
          let ro = solve subst1 in
          let sigma = solve [(typ1, Boolean)] in
          let (typ2, _, subst2) = gts e1 ((applySubstEnv (ro @ sigma) (fst o)), snd o) in
          let roP = solve subst2 in
          let sigmaP = solve [(typ2, Boolean)] in
          (Boolean, o, ro @ sigma @ roP @ sigmaP)
      
      | Unit -> (UnitType, o, [])
      | Native (_, _, t) -> (t, o, [])
      | Dummy -> (DummyType, o, [])
      | Unknown _ -> raise (Type_error "Unexpected unknown")

(* gtsd = Get Type and Substitution in a Declaration 
This function takes a declaration d and a type schema environment o and 
returns a three-element tuple: the declaration's type if it has one,
the new type schema environment and the substitution solution of the type 
equations generated by type inference.
Raises Type_error if d is not well-typed *)
and gtsd (d : declaration) (o : envType * typedefs) =
  match d with
    | Def (f, args, e) ->
        gtsd (DefType (f, Null, args, List.map (fun _ -> Null) args, e)) o
    | Defrec (f, args, e) ->  
        gtsd (DefrecType (f, Null, args, List.map (fun _ -> Null) args, e)) o
    | DefType (f, typef, [], typeargs, e) -> 
        let (t, _, subst) = gts e o in
        let ro = solve subst in
        if typef = Null
        then (t, o, ro)
        else     
           let sigma = solve [(typef, t)] in 
           (typef, o, ro @ sigma)
    | DefType (f, typef, args, typeargs, e) -> 
        let (t, _, subst) = gts (FunType (args, typeargs, e)) o in
        let ro = solve subst in
        if typef = Null
        then (t, o, ro)
        else
          let sigma = solve [(typef, getReturnType t (List.length args))] in
          (applySubst sigma t, o, ro @ sigma)
    | DefrecType (f, Null, [], typeargs, e) -> 
        let freshVar = newVar() in
        let oP = Environment_types.bind f (Schema([], freshVar)) (fst o) in
        let (t, _, subst) = gts e (oP, snd o) in
        let ro = solve subst in
        let aux = applySubst ro freshVar in
        let sigma = solve [(t, aux)] in
        (applySubst sigma t, o, ro @ sigma)
    | DefrecType (f, typef, [], typeargs, e) -> 
        let oP = Environment_types.bind f (Schema([], typef)) (fst o)  in
        let (t, _, subst) = gts e (oP, snd o) in
        let ro = solve subst in
        let sigma = solve [(typef, t)] in  
        (typef, o, ro @ sigma)       
    | DefrecType (f, typef, [x], typeargs, e) -> 
        let freshVar = newVar() in
        let s = if typef = Null then newVar() else typef in
        let oP = Environment_types.bind f (Schema ([], Arrow (freshVar, s))) (fst o) in
        let (t, _, subst) = gts (FunType ([x], typeargs, e)) (oP, snd o) in
        let ro = solve subst in
        let aux = applySubst ro (Arrow (freshVar, s)) in
        let sigma = solve [(t, aux)] in
        (applySubst sigma t, o, ro @ sigma)
    | DefrecType (f, typef, a :: t, typeargs, e) -> 
        let freshVar = newVar () in
        let s = if typef = Null then newVar () else typef in
        gtsd (DefrecType (f, Arrow (freshVar, s), [a], [List.hd typeargs], FunType (t, List.tl typeargs, e))) o
    | Defcorec(_, f, args, e) -> gtsd (Defrec(f, args, e)) o
    | DefcorecType(_, f, typef, args, typeargs, e) ->
        gtsd (DefrecType (f, typef, args, typeargs, e)) o
    | Typedef _ -> assert false (* never called on this; would not make sense *)

let getType (t : expr) (o: envType * typedefs) : typeNC * envType * typedefs =
  let (t, (s, td), _) = gts t o in (t, s, td)

let getTypeDecl (d : declaration) (o : envType * typedefs) : typeNC * envType * typedefs =
  let (t, (s, td), _) = gtsd d o in (t, s, td)
